# {DirTravel} is a library for getting information about files and
# directories recursively to a tree structure. This library extends
# the RubyTree classes to include directory and files info. Please
# refer to RubyTree documentation for RubyTree related features.
#
# A proxy object {DirTravel::Travel} is used to gather the directory
# content. The "filetree" class method provides interface for getting
# the content with few options (see: {DirTravel::Travel.filetree} for
# details).
#
# See the {DirTravel::Entry} methods doc for possibilities in
# examining the directory hierachy.
#
#
# Examples:
#   require 'dirtravel'
#
#   # Collect all entries under '.'
#   d = DirTravel::Travel.filetree( '.' )
#
#   # Display names of the entries in the hierarhcy.
#   d.children.each do |i|
#     puts i.name
#   end
#
#   # Display absolute path for file (leaf level) entries.
#   d.each_leaf do |i|
#     puts i.abspath
#   end
#
#   # Get MP3 files and create a list of all album directories.
#   #   Assume: ".../<album>/<song>" hierarhcy
#   #
#   d = DirTravel::Travel.filetree( '.', { :suffix => '.mp3' } )
#   albums = d.select_level( d.node_height - 1 )
#

module DirTravel

    require 'rubytree'


    # {DirTravel} error.
    class DirTravelError < RuntimeError; end


    # Extend RubyTree base class with file and directory features.
    class Entry < Tree::TreeNode

        # Node name.
        attr_accessor :name

        # Set name for {Entry} (Dir/File).
        # Initialize abspath.
        def initialize( name )
            super( name, nil )
            @abspath = nil
        end


        # Return path components as Array.
        #
        # @param basedir [Entry] Starting level for the hierarchy.
        # @return [Array] Array of names in hierarchy.
        def parts( basedir = self )
            parents = []
            while basedir
                parents.push basedir.tip
                basedir = basedir.parent
            end
            parents.reverse
        end

        alias pathArray parts


        # Relative path.
        def path
            parts.join( '/' )
        end


        # Relative path under root.
        #
        # @param level [Integer] Level down from root.
        def subpath( level = 1 )
            pa = parts
            if level < 0 || level > pa.length
                raise DirTravelError, "Invalid index for subpath level!"
            end
            pa[ level .. -1 ].join( '/' )
        end


        # Absolute path.
        def abspath
            if @abspath
                @abspath
            else
                root.abspath + '/' + subpath
            end
        end


        # Top directory name (usually same as name).
        def tip
            @name.split( "/" )[-1]
        end


        # Relative path of parenting directory.
        #
        # @param basedir [Entry] Starting level for the hierarchy.
        # @return [String] Containing directory.
        def dir( basedir = self )
            parts( basedir.parent ).join( '/' )
        end


        # Select all siblings from given node depth.
        #
        # @param level [Integer] Selected level in hierachy. Level is
        #   number of steps down in hierarhcy.
        # @return [Array] Array of siblings in selected hierarchy.
        def select_level( level )
            select do |i| i.node_depth == level; end
        end


        # Return all file entries in hierarchy.
        def files
            select do |i| i.kind_of?( FileEntry ) end
        end


        # File.stat data for the {Entry}.
        def stat
            File.stat( path )
        end


        # Relative path {Entry}.
        def relative?
            @name[0] != '/'
        end


        # Rename node.
        #
        # @param name [String] If name is abspath then {Entry} becomes
        #                      abspath.
        def rename( name )
            @name = name

            # Absolute or relative path?
            if name[0] == "/"
                @abspath = name
            else
                @abspath = nil
            end
        end

    end


    # Directory type entry.
    class DirEntry < Entry

        # Set name and abspath if given.
        #
        # @param name [String] Directory name.
        # @param abspath [String] Set abspath if given. Otherwise it
        #                         is generated from pieces.
        def initialize( name, abspath = nil )
            super( name )
            if abspath
                @abspath = abspath
            end
        end

    end


    # File type entry.
    class FileEntry < Entry
        attr_reader :suffix, :basename

        # Set name, suffix, and basename.
        #
        # @param name [String] File name.
        def initialize( name )
            super( name )
            @suffix = File.extname( name )
            @basename = File.basename( name, @suffix )
        end

    end


    # Create directory recursion tree (with
    # {Travel.filetree}). Optionally filter with suffix and modify tree
    # building with options Hash (see below).
    #
    # If basedir is absolute path, the root name is the search
    # path. If basedir is relative path, before search the current
    # directory is changed to referenced directory and root name is
    # the current directory, single level.
    #
    # == Parameters:
    # sort::
    #   Sort directory entries (default: false).
    # suffix::
    #   Limit search to files with "suffix" (default: nil).
    # files::
    #   Include files to search (default: true).
    # inclusive::
    #   Basedirs parent becomes the basedir (default: false).
    #
    # Example:
    #    d1 = DirTravel::Travel.filetree( dir1, { :sort => true, :suffix => '.mp3' } )

    class Travel

        # Root {DirEntry} of {Travel}.
        attr_accessor :root

        # Default options for {Travel}.
        attr_accessor :defaults

        # Starting directory for {Travel}.
        attr_accessor :basedir
        attr_accessor :abspath


        # Create directory recursion tree and return root.
        #
        # @param basedir [String] Starting directory (top).
        # @param options [Hash] {Travel} options.
        # @return [DirEntry] Root of the file travel hierarchy.
        def Travel.filetree( basedir = '.', options = {} )

            # Non-nil if directory needs to be changed.
            pwd = nil

            if basedir[0] == '/'

                # Absolue path.
                t = Travel.new( basedir, basedir, options )
                t.travel

            else

                # Relative path.

                # Store current directory before chdir.
                pwd = Dir.pwd

                # Generate target directory for chdir. One up from
                # the reference.
                full = File.absolute_path( basedir )
                base = File.basename( full )
                dir = File.dirname( full )

                # Goto target.
                Dir.chdir( dir )

                t = Travel.new( base, full, options )
                t.travel
            end


            if t.defaults[ :inclusive ]

                # With inclusive the root is changed to one-up from
                # target.

                # One up from root.
                uppath = File.dirname( t.root.abspath )
                if t.root.relative?
                    path = File.basename( uppath )
                else
                    path = uppath
                end

                # Create the "one-up" root.
                newRoot = DirEntry.new( path, uppath )

                # Rename old root.
                t.root.rename( t.root.tip )

                # Add to one-up root.
                newRoot.add( t.root )

                # Set root to one-up root.
                t.root = newRoot
            end

            # Return back to start directory if dir changed.
            Dir.chdir( pwd ) if pwd

            # Return root {DirEntry}.
            t.root
        end


        # Create directory recursion object. Overlay options on top of
        # defaults. Initialize root {DirEntry}.
        #
        # @param basedir [String] Starting directory (top).
        # @param abspath [String] Absolute path for root.
        # @param options [Hash] {Travel} options.
        def initialize( basedir, abspath, options = {} )
            @basedir = basedir
            @abspath = abspath

            @defaults = {
                :suffix    => nil,
                :sort      => false,
                :files     => true,
                :inclusive => false,
            }

            @defaults.merge!( options )

            @root = DirEntry.new( basedir, abspath )
        end


        # Recursively get all files with suffix. Ignore suffix if
        # suffix is nil.
        def travel( suffix = @defaults[ :suffix ] )
            entriesIn( @basedir, @root, suffix )
        end


        private

        # Recursively get all files with suffix. Ignore suffix if
        # suffix is nil.
        def entriesIn( dir, node, suffix = nil )

            list = entries( dir )

            list.each do |i|

                if File.file?( dir + '/' + i )

                    # File entry.

                    if ( !suffix || suffix == File.extname( i ) ) &&
                            @defaults[ :files ]
                        node.add( FileEntry.new( i ) )
                    end

                else

                    # Dir entry.

                    newNode = DirEntry.new( i )
                    node.add( newNode )
                    entriesIn( dir + '/' + i, newNode, suffix )

                end

            end

            self
        end

        # Directory entries with dot entries filtered out.
        def entries( dir )
            entries = Dir.entries( dir ).select do |i| nonDotEntry( i ); end
            entries.sort! if @defaults[ :sort ]
            entries
        end

        # Test for non-dot-entry.
        def nonDotEntry( name )
            ( name != '.' && name != '..' )
        end

    end


end


require_relative 'version'
